#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/ingest/T0Compiler.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 11.05.2021
# Last Modified Date: 11.05.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, List, Set, Union, Tuple, Any
from ampel.content.MetaRecord import MetaRecord
from ampel.types import ChannelId, DataPointId
from ampel.content.DataPoint import DataPoint
from ampel.abstract.AbsDocIngester import AbsDocIngester
from ampel.util.collections import try_reduce
from ampel.abstract.AbsCompiler import AbsCompiler


class T0Compiler(AbsCompiler):
	"""
	Compiles updates to t0 documents arising from different t0 directives
	"""

	def __init__(self, **kwargs) -> None:
		super().__init__(**kwargs)
		self.register: Dict[int, List[Tuple[Set[ChannelId], List[DataPoint]]]] = {}


	# Override
	def add(self, # type: ignore[override]
		dps: List[DataPoint],
		channel: ChannelId,
		trace_id: int
	) -> None:

		if trace_id not in self.register:
			self.register[trace_id] = []
		for chans, dps_to_ingest in self.register[trace_id]:
			if dps == dps_to_ingest:
				chans.add(channel)
				return
		self.register[trace_id].append(({channel}, dps))


	# Override
	def commit(self, ingester: AbsDocIngester[DataPoint], now: Union[int, float]) -> None:
		"""
		Note that we let the ingester handle 'ts' and 'updated' values
		"""

		x: Dict[DataPointId, DataPoint] = {}
		for trace_id, channel_sets in self.register.items():

			for chans, dps in channel_sets:

				for dp in dps:

					if self._tag and 'tag' not in dp:
						dp['tag'] = self._tag

					if self.origin and 'origin' not in dp:
						dp['origin'] = self.origin

					meta: MetaRecord = {
						'run': self.run_id,
						'ts': now,
						'channel': try_reduce(list(chans)),
						'traceid': {'shaper': trace_id}
					}

					if 'meta' in dp:
						dp['meta'].append(meta) # type: ignore[attr-defined]
					else:
						dp['meta'] = [meta] # type: ignore[list-item]

					if 'channel' in dp:
						if isinstance(dp['channel'], list):
							dp['channel'] = chans | set(dp['channel']) # type: ignore[typeddict-item]

						else:
							dp['channel'] |= chans # type: ignore[operator]
					else:
						dp['channel'] = chans # type: ignore[typeddict-item]

					if dp['id'] not in x:
						# add a new point
						x[dp['id']] = dp
					else:
						# update channel set and metadata
						prev = x[dp['id']]
						prev['channel'] |= dp['channel'] # type: ignore[operator]
						if isinstance(pchan := (meta := prev['meta'][-1])['channel'], list):
							prev_chans = set(pchan)
						else:
							prev_chans = {pchan}
						meta['channel'] = try_reduce(list(prev_chans | chans))

		for dp in x.values():
			dp['channel'] = list(dp['channel'])
			ingester.ingest(dp, now)

		self.register.clear()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/abstract/AbsT3Stager.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 04.01.2020
# Last Modified Date: 18.04.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Union, List, Dict, Any, Optional, Generator

from ampel.types import ChannelId
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.content.T3Document import T3Document
from ampel.core.ContextUnit import ContextUnit
from ampel.mongo.update.MongoStockUpdater import MongoStockUpdater
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.log.AmpelLogger import AmpelLogger


class AbsT3Stager(AmpelABC, ContextUnit, abstract=True):
	"""
	Supply stock views to one or more T3 units.
	"""

	logger: AmpelLogger
	stock_updr: MongoStockUpdater

	channel: Optional[ChannelId] = None

	#: raise exceptions instead of catching and logging
	raise_exc: bool = True

	#: contextual information for this run
	session_info: Optional[Dict[str, Any]] = None

	#: number of buffers to process at once. Set to 0 to disable chunking
	chunk_size: int = 1000

	#: Cast ampel buffers into views for each t3 unit (meaning possibly redundantly)
	#: since there is no real read-only struct in python
	paranoia: bool = True


	@abstractmethod
	def stage(self, data: Generator[AmpelBuffer, None, None]) -> Optional[Union[T3Document, List[T3Document]]]:
		""" Process a chunk of AmpelBuffer instances """



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/mongo/update/MongoStockIngester.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 14.12.2017
# Last Modified Date: 24.05.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Any, Union
from pymongo import UpdateOne
from ampel.abstract.AbsDocIngester import AbsDocIngester
from ampel.content.StockDocument import StockDocument
from ampel.mongo.utils import maybe_use_each
from ampel.util.collections import try_reduce


class MongoStockIngester(AbsDocIngester[StockDocument]):

	def ingest(self, doc: StockDocument, now: Union[int, float]) -> None:

		set_on_insert: dict[str,Any] = {'ts.any.tied': now}
		add_to_set = {'channel': maybe_use_each(doc['channel'])}

		if 'name' in doc:
			set_on_insert['name'] = doc['name']
			
		if 'tag' in doc:
			add_to_set['tag'] = maybe_use_each(doc['tag'])

		for el in doc['journal']:
			el['channel'] = try_reduce(el['channel'])

		upd = {
			'$addToSet': add_to_set,
			'$min': {f'ts.{chan}.tied': now for chan in doc['channel']},
			'$max': {f'ts.{chan}.upd': now for chan in doc['channel']},
			'$push': {'journal': maybe_use_each(doc['journal'])}
		}

		if set_on_insert:
			upd['$setOnInsert'] = set_on_insert

		# Insert/Update stock document into stock collection
		self.updates_buffer.add_stock_update(
			UpdateOne(
				{'stock': doc['stock'], 'origin': doc['origin']}
				if 'origin' in doc else {'stock': doc['stock']},
				upd,
				upsert=True
			)
		)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/abstract/AbsCompiler.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 07.05.2021
# Last Modified Date: 27.05.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Optional, Literal, Sequence, Union
from ampel.types import Tag
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.abstract.AbsDocIngester import AbsDocIngester


class AbsCompiler(AmpelABC, AmpelBaseModel, abstract=True):

	origin: Optional[int] = None
	tier: Literal[-1, 0, 1, 2, 3]
	run_id: int
	tag: Optional[Union[Tag, Sequence[Tag]]]


	def __init__(self, **kwargs) -> None:
		super().__init__(**kwargs)
		self._tag = None
		if self.tag:
			if isinstance(self.tag, (str, int)):
				self._tag = [self.tag]
			else:
				self._tag = list(self.tag)


	@abstractmethod(var_args=True)
	def add(self) -> None:
		...

	@abstractmethod
	def commit(self, ingester: AbsDocIngester, now: Union[int, float]) -> None:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/ingest/T2Compiler.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 01.01.2018
# Last Modified Date: 06.09.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from ujson import encode, decode
from typing import Optional, Dict, Union, Tuple, Set, Any
from ampel.types import ChannelId, UnitId, T2Link, StockId
from ampel.content.T2Document import T2Document
from ampel.enum.MetaActionCode import MetaActionCode
from ampel.abstract.AbsDocIngester import AbsDocIngester
from ampel.abstract.AbsCompiler import AbsCompiler
from ampel.util.collections import try_reduce


class T2Compiler(AbsCompiler):
	"""
	Helps build a minimal set of :class:`compounds <ampel.content.T2Document.T2Document>`
	Multiple T2 documents associated with the same stock record accross different channels
	are merged into one single T2 document that references all corresponding channels.
	"""

	def __init__(self, col: Optional[str] = None, **kwargs) -> None:
		super().__init__(**kwargs)
		self.t2s: Dict[
			# key: (unit name, unit config, link, stock)
			Tuple[UnitId, Optional[int], T2Link, StockId],
			# tuple[channels, dict[serialized(meta)|None, {channels}]]
			Tuple[Set[ChannelId], Dict[Optional[str], Set[ChannelId]]]
		] = {}
		self.col = col


	def add(self, # type: ignore[override]
		unit: UnitId,
		config: Optional[int],
		stock: StockId,
		link: T2Link,
		channel: ChannelId,
		meta: Optional[Dict[str, Any]] = None
	) -> None:

		# unprocessed T1 documents requiring processing have stock equals zero
		if stock == 0:
			return

		k = (unit, config, link, stock)
		k2 = encode(meta, sort_keys=True) if meta else None
		
		d = self.t2s.get(k)
		if d:
			d[0].add(channel)
			if k2 in d[1]:
				d[1][k2].add(channel)
			else:
				d[1][k2] = {channel}
		else:
			self.t2s[k] = {channel}, {k2: {channel}}


	def commit(self, ingester: AbsDocIngester[T2Document], now: Union[int, float]) -> None:

		for k, v in self.t2s.items():

			# Note: mongodb maintains key order
			d: T2Document = {'unit': k[0], 'config': k[1], 'link': k[2]}

			if self.col:
				d['col'] = self.col

			d['stock'] = k[3]

			if self.origin:
				d['origin'] = self.origin

			d['channel'] = list(v[0])

			if self._tag:
				d['tag'] = self._tag

			d['meta'] = []
			for k2, v2 in v[1].items():

				entry = {
					'run': self.run_id,
					'ts': now,
					'tier': self.tier,
					'channel': try_reduce(list(v2)),
					'action': MetaActionCode.ADD_CHANNEL
				}

				if self._tag:
					entry['action'] |= MetaActionCode.ADD_TAG
					entry['tag'] = try_reduce(self._tag)

				if k2 and (x := decode(k2)):
					entry |= x

				d['meta'].append(entry) # type: ignore[attr-defined]

			ingester.ingest(d, now)

		self.t2s.clear()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/t3/supply/SimpleT2BasedSupplier.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 15.07.2021
# Last Modified Date: 15.07.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Generator, Dict, Any
from ampel.abstract.AbsT3Supplier import AbsT3Supplier
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.log.utils import safe_query_dict


class SimpleT2BasedSupplier(AbsT3Supplier):

	query: Dict[str, Any]

	def supply(self) -> Generator[AmpelBuffer, None, None]:

		if self.logger.verbose > 1: # log query parameters
			self.logger.debug(
				None, extra={
					'col': 2,
					'query': safe_query_dict(self.query, dict_key=None)
				}
			)

		# Retrieve pymongo cursor
		col = self.context.db.get_collection("t2")

		d: Dict[int, AmpelBuffer] = {}
		for el in col.find(self.query):
			if el['stock'] in d:
				d[el['stock']]['t2'].append(el) # type: ignore
			else:
				d[el['stock']] = AmpelBuffer(
					id = el['stock'],
					t2 = [el]
				)

		for ab in d.values():
			yield ab



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/cli/AbsCoreCommand.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 18.03.2021
# Last Modified Date: 15.07.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

import re
from typing import Sequence, Dict, Any, Optional, TypeVar, Type
from ampel.core.AmpelDB import AmpelDB
from ampel.config.AmpelConfig import AmpelConfig
from ampel.abstract.AbsCLIOperation import AbsCLIOperation
from ampel.core.AmpelContext import AmpelContext
from ampel.core.UnitLoader import UnitLoader
from ampel.log.AmpelLogger import AmpelLogger
from ampel.util.mappings import set_by_path
from ampel.util.freeze import recursive_freeze

custom_conf_patter = re.compile(r"^--[\w-]*(?:\.[\w-]+)*.*$")

T = TypeVar("T", bound = AmpelContext)

class AbsCoreCommand(AbsCLIOperation, abstract=True):


	def load_config(self,
		config_path: str,
		customizations: Sequence[str],
		logger: Optional[AmpelLogger] = None,
		freeze: bool = True
	) -> AmpelConfig:

		ampel_conf = AmpelConfig.load(config_path, freeze=False)

		if logger is None:
			logger = AmpelLogger.get_logger()

		it = iter(customizations)
		for el in it:
			if custom_conf_patter.match(el):
				if "=" in el:
					s = el.split("=")
					k = s[0][2:]
					v = _maybe_int(s[1])
				else:
					k = el[2:]
					try:
						v = _maybe_int(next(it))
					except StopIteration:
						return ampel_conf # TODO: say something ?

				if ampel_conf.get(".".join(k.split(".")[:-1])) is None:
					logger.info(f"Cannot set ampel config parameter '{k}' (parent key does not exist)")
					continue

				logger.info(f"Setting config parameter '{k}' value to: {v}")
				set_by_path(ampel_conf._config, k, v)

		if freeze:
			ampel_conf._config = recursive_freeze(ampel_conf._config)

		return ampel_conf


	def get_context(self,
		args: Dict[str, Any],
		unknown_args: Sequence[str],
		logger: Optional[AmpelLogger] = None,
		freeze_config: bool = True,
		ContextClass: Type[T] = AmpelContext, # type: ignore[assignment]
		**kwargs
	) -> T:

		config = self.load_config(
			args['config'], unknown_args,
			logger or AmpelLogger.get_logger(),
			freeze = freeze_config
		)

		vault = None
		if args.get('vault'):
			from ampel.secret.AmpelVault import AmpelVault
			from ampel.secret.DictSecretProvider import DictSecretProvider
			vault = AmpelVault([DictSecretProvider.load(args['secrets'])])

		db = AmpelDB.new(config, vault)
		return ContextClass(
			config = config,
			db = db,
			loader = UnitLoader(config, db=db, vault=vault),
			**kwargs
		)


def _maybe_int(stringy):
	try:
		return int(stringy)
	except Exception:
		return stringy



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/core/ContextUnit.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 07.10.2019
# Last Modified Date: 18.06.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.core.AmpelContext import AmpelContext
from ampel.log.AmpelLogger import AmpelLogger
from ampel.abstract.Secret import Secret
from ampel.util.mappings import dictify


class ContextUnit(AmpelBaseModel):
	"""
	Base class for units requiring a reference to an AmpelContext instance
	"""

	#: Private variable potentially set by UnitLoader for provenance purposes. Either:
	#: * 0 if provanance flag is False
	#: * -1 in case model content is not serializable
	#: * any other signed int value
	_trace_id: int = 0

	def __init__(self, context: AmpelContext, **kwargs) -> None:

		if context is None:
			raise ValueError("Parameter context cannot be None")

		super().__init__(**kwargs)

		self._trace_content = {
			k: dictify(v)
			for k, v in self.__dict__.items()
			if not isinstance(v, (Secret, AmpelContext, AmpelLogger))
		}

		self.context = context



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/mongo/view/MongoChannelView.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 26.03.2021
# Last Modified Date: 26.03.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import List, Any, Dict
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.base.AmpelBaseModel import AmpelBaseModel


class AbsMongoView(AmpelABC, AmpelBaseModel, abstract=True):

	@abstractmethod
	def stock(self) -> List[Dict[str, Any]]:
		...

	@abstractmethod
	def t0(self) -> List[Dict[str, Any]]:
		...

	@abstractmethod
	def t1(self) -> List[Dict[str, Any]]:
		...

	@abstractmethod
	def t2(self) -> List[Dict[str, Any]]:
		...

	@abstractmethod
	def t3(self) -> List[Dict[str, Any]]:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/mongo/view/AbsMongoFlatMultiView.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 26.03.2021
# Last Modified Date: 28.03.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import List, Any, Dict, Sequence
from ampel.types import ChannelId
from ampel.base.decorator import abstractmethod
from ampel.mongo.view.AbsMongoView import AbsMongoView


class AbsMongoFlatMultiView(AbsMongoView, abstract=True):
	"""
	Handles flat AND or OR connected type of channel-based views (any number of)
	No support for nested logic schema (such as {any_of: [A, all_of: {[B, C]}]})
	hence the 'flat' in class name
	"""

	channel: Sequence[ChannelId]
	t0_has_chan: bool = True


	@abstractmethod
	def get_first_match(self) -> Dict[str, Any]:
		...


	def stock(self) -> List[Dict[str, Any]]:

		return [
			self.get_first_match(),
			{
				'$project': {
					'tag': 1,
					'name': 1,
					'body': 1,
					'channel': {'$setIntersection': [self.channel, '$channel']},
					'journal': self.morph_seq('journal'),
					'ts': {chan: f'$ts.{chan}' for chan in self.channel},
				}
			}
		]


	def t0(self) -> List[Dict[str, Any]]:

		ret = [
			{
				'$set': {
					'channel': self.get_channel_intersection(),
					'meta': self.morph_seq("meta")
				}
			}
		]

		if self.t0_has_chan:
			ret.insert(0, self.get_first_match())

		return ret


	def t1(self) -> List[Dict[str, Any]]:

		return [
			self.get_first_match(),
			{
				'$set': {
					'channel': self.get_channel_intersection(),
					'meta': self.morph_seq("meta")
				}
			}
		]


	def t2(self) -> List[Dict[str, Any]]:

		return [
			self.get_first_match(),
			{
				'$set': {
					'channel': self.get_channel_intersection(),
					'meta': self.morph_seq("meta")
				}
			}
		]


	def t3(self) -> List[Dict[str, Any]]:

		return [
			self.get_first_match(),
			{
				'$set': {
					'channel': self.get_channel_intersection(),
					'meta': self.morph_seq("meta")
				}
			}
		]


	def morph_seq(self, arg: str) -> Dict[str, Any]:
		"""
		If channel is an array, reduce its value to the intersection between
		the defined channels and the requested one (should work with AND or OR channel projections).
		"""

		or_list = [
			{'$eq': [chan, '$$d.channel']}
			for chan in self.channel
		]

		or_list.insert(0, {"$not": ["$$d.channel"]})

		return {
			'$filter': {
				'input': {
					'$map': {
						'input': f"${arg}",
						'as': "d",
						'in': {
							'$cond': {
								'if': {'$isArray': '$$d.channel'},
								'then': {
									'$let': {
										'vars': {
											'intersect': {'$setIntersection': [self.channel, '$$d.channel']},
										},
										'in': {
											'$switch': {
												'branches': [
													{
														'case': {'$eq': [{'$size': '$$intersect'}, 0]},
														'then': '$$REMOVE'
													},
													{
														'case': {'$eq': [{'$size': '$$intersect'}, 1]},
														'then': {'$mergeObjects': ['$$d', {'channel': {"$arrayElemAt": ['$$intersect', 0]}}]}
													},
													{
														'case': {'$gt': [{'$size': '$$intersect'}, 1]},
														'then': {'$mergeObjects': ['$$d', {'channel': '$$intersect'}]}
													}
												]
											}
										}
									}
								},
								'else': {
									'$cond': {
										'if': {'$or': or_list},
										'then': '$$d',
										'else': '$$REMOVE'
									},
								},
							}
						}
					}
				},
				'as': 'd',
				'cond': {'$ne': ['$$d', None]}
			}
		}


	def get_channel_intersection(self) -> Any:

		return {
			'$let': {
				'vars': {
					'intersect': {'$setIntersection': [self.channel, '$channel']},
				},
				'in': {
					'$cond': {
						'if': {'$eq': [{'$size': '$$intersect'}, 1]},
						'then': {"$arrayElemAt": ['$$intersect', 0]},
						'else': '$$intersect'
					},
				}
			}
		}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/abstract/AbsDocUpdater.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 28.05.2021
# Last Modified Date: 28.05.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from typing import Dict, Any, Literal
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.mongo.update.DBUpdatesBuffer import DBUpdatesBuffer


class AbsDocUpdater(AmpelABC, AmpelBaseModel, abstract=True):
	"""
	Aim at facilating updates of documents regardless of the underlying database system.
	Not in use yet
	"""

	tier: Literal[0, 1, 2, 3]
	updates_buffer: DBUpdatesBuffer

	@abstractmethod
	def update(self, match: Dict[str, Any], set: Dict[str, Any], push: Dict[str, Any], add_to_set: Dict[str, Any]) -> None:
		...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/cli/ConfigCommand.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 17.07.2021
# Last Modified Date: 31.08.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>

from argparse import ArgumentParser
from typing import Sequence, Dict, Any, Optional, Union
from ampel.cli.AbsCoreCommand import AbsCoreCommand
from ampel.cli.AmpelArgumentParser import AmpelArgumentParser
from ampel.cli.ArgParserBuilder import ArgParserBuilder
from ampel.config.builder.DistConfigBuilder import DistConfigBuilder


hlp = {
	'config': 'Path to an ampel config file (yaml/json)',
	# Optional
	'secrets': 'Path to a YAML secrets store in sops format',
	'out': 'Path to file where config will be saved (printed to stdout otherwise)',
	'sign': 'Append truncated file signature (last 6 digits) to filename',
	'stop-on-errors': 'by default, config building stops and raises an exception if an error occured.\n' +
		'- 2: stop on errors\n' +
		'- 1: ignore errors in first_pass_config only (will stop on morphing/scoping/template errors)\n' +
		'- 0: ignore all errors',
	'verbose': 'verbose'
}


class ConfigCommand(AbsCoreCommand):
	"""
	"""

	def __init__(self):
		self.parsers = {}


	# Mandatory implementation
	def get_parser(self, sub_op: Optional[str] = None) -> Union[ArgumentParser, AmpelArgumentParser]:

		if sub_op in self.parsers:
			return self.parsers[sub_op]

		sub_ops = ['build', 'show']

		if sub_op is None or sub_op not in sub_ops:
			return AmpelArgumentParser.build_choice_help(
				'config', sub_ops, hlp, description = 'Build or show ampel config.'
			)

		builder = ArgParserBuilder("config")
		builder.add_parsers(sub_ops, hlp)

		# Required args
		builder.add_arg("build.required", "out")

		# Optional args
		builder.add_arg("optional", "secrets", default=None)
		builder.add_arg('optional', 'verbose', action="store_true")
		builder.add_arg("optional", "log-profile", default="default")

		builder.add_arg("build.optional", "sign", action="store_true")
		builder.add_arg("show.optional", "pretty", action="store_true")
		builder.add_arg("build.optional", "stop-on-errors", default=2)

		# Example
		builder.add_example("build", "-out ampel_conf.yaml")
		builder.add_example("build", "-out ampel_conf.yaml -sign -verbose")
		builder.add_example("show", "-pretty")

		self.parsers.update(
			builder.get()
		)

		return self.parsers[sub_op]



	# Mandatory implementation
	def run(self, args: Dict[str, Any], unknown_args: Sequence[str], sub_op: Optional[str] = None) -> None:

		cb = DistConfigBuilder(
			verbose = args.get('verbose', False)
		)

		cb.load_distributions()

		cb.build_config(
			stop_on_errors = 0,
			skip_default_processes=True,
			config_validator = None,
			save = args['out']
		)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File              : Ampel-core/ampel/util/getch.py
# License           : BSD-3-Clause
# Author            : vb <vbrinnel@physik.hu-berlin.de>
# Date              : 24.06.2021
# Last Modified Date: 24.06.2021
# Last Modified By  : vb <vbrinnel@physik.hu-berlin.de>


class _Getch:
	"""
	Gets a single character from standard input.
	Inspired by https://stackoverflow.com/questions/510357/how-to-read-a-single-character-from-the-user
	"""
	def __init__(self, impl) -> None:
		self.impl = impl

	def __call__(self) -> bool:
		char = self.impl()
		# CTRL-C or q or ESC
		if char == '\x03' or char == 'q' or ord(char) == 27:
			return True
		elif char == '\x04':
			raise EOFError
		return False


class _GetchUnix:

	def __call__(self):
		fd = sys.stdin.fileno()
		old_settings = termios.tcgetattr(fd)
		try:
			tty.setraw(sys.stdin.fileno())
			ch = sys.stdin.read(1)
		finally:
			termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
		return ch


class _GetchWindows:

	def __call__(self):
		return msvcrt.getch()


try:
	import msvcrt
	getch = _Getch(_GetchWindows())
except ImportError:
	import sys, tty, termios
	getch = _Getch(_GetchUnix())
